/*
 * RT-Linux scheduler
 *
 * Written by Michael Barabanov, Victor Yodaiken
 * Copyright (C) VJY Associates LLC, 1998-1999
 * Released under the terms of the LGPL
 *
 */

#ifndef __RTL__SCHED__
#define __RTL__SCHED__

#include <rtl_time.h>
#include <asm/spinlock.h>
#include <asm/ptrace.h>
#include <linux/tasks.h>
#include <linux/rtl.h>
#include <linux/tqueue.h>
#include "schedulers/x86/fpu.h"
/* #include "../schedulers/x86/switch.h" */

#define RTL_DEBUG_HEARTBEAT	1
#define RTL_DEBUG_TIMER	 	2
extern int rtl_debug (int state);

enum rtl_task_states {
	RTL_THREAD_READY, RTL_THREAD_DELAYED, RTL_THREAD_DORMANT, RTL_THREAD_ACTIVE, RTL_THREAD_ZOMBIE
};

/* the only available policy so far is SCHED_FIFO */
#define SCHED_OTHER             0
#define SCHED_FIFO              1
#define SCHED_RR                2

/* struct sched_param is defined in linux/sched.h, so we use rtl_sched_param */
struct rtl_sched_param {
	int sched_priority; /* note that larger values represent higher priorities */
};


#define TIMER_ABSTIME 1
#define RTL_MAX_TIMERS 32

struct rtl_timer_struct {
	struct itimerspec spec;
	clockid_t clock;
	char pad[16];
};


typedef struct rtl_timer_struct *timer_t;

struct rtl_thread_struct {
	int *stack;	/* hardcoded */
	int uses_fp;
	enum rtl_task_states state;
	int *stack_bottom;
	struct rtl_sched_param sched_param;
	struct rtl_thread_struct *next;
	RTL_FPU_CONTEXT fpu_regs;
	int cpu;
	timer_t timer;
	void *retval;
	int being_destroyed;
	struct tq_struct free_task;
	struct tq_struct move_task;
	int pad [64];
};


extern void rtl_schedule (void);

typedef struct rtl_thread_struct RTL_THREAD;

typedef RTL_THREAD *pthread_t;
 

#define rtl_is_linux_task(cpu, a) (a == &sched_data(cpu)->rtl_linux_task)

struct rtl_sched_cpu_struct {
	spinlock_t sched_lock;
	RTL_THREAD *rtl_current;
	RTL_THREAD rtl_linux_task;
	RTL_THREAD idle_task;
	RTL_THREAD *rtl_task_fpu_owner;	/* the task whose FP context is currently in the FPU unit */
	int rtl_periodic_mode;
	int rtl_period;
	RTL_THREAD *rtl_tasks;
	task_queue rtl_tq_cpu;
/* 	char pad[64]; */
} __attribute__ ((aligned (64)));


extern struct rtl_sched_cpu_struct rtl_sched [NR_CPUS];

#define sched_data(cpu) (&rtl_sched [cpu])

/* RTL-specific function TODO: write POSIX equivalents for these */

extern int rtl_pthread_wait_timer (timer_t timer);

extern int rtl_pthread_delete (pthread_t thread);

extern int rtl_pthread_assign_to_cpu (pthread_t thread, int cpu);

extern int rtl_pthread_use_fp (pthread_t thread, int flag);

extern int rtl_pthread_suspend (pthread_t thread);

extern int rtl_pthread_wakeup (pthread_t thread);

extern int rtl_clock_set_periodic_mode (clockid_t clock, struct timespec *period);

extern int rtl_clock_set_oneshot_mode (clockid_t clock);

/* end RTL-specific */


/* POSIX interface */

extern int timer_create(clockid_t clockid, void /*struct sigevent*/ *evp, timer_t *timerid);
extern int timer_delete(timer_t timerid);
extern int timer_settime(timer_t timerid, int flags,
		    const struct itimerspec *value, struct itimerspec *ovalue);
extern int timer_gettime(timer_t timerid, struct itimerspec *value);
extern int timer_getoverrun(timer_t timerid);

extern inline int sched_get_priority_max(int policy) { return 1000000; }
extern inline int sched_get_priority_min(int policy) { return 0; }

extern inline pthread_t pthread_self(void) {
	return rtl_sched [rtl_get_cpu_id()] . rtl_current;
}

typedef struct STRUCT_PTHREAD_ATTR {
	int stack_size;
	struct rtl_sched_param sched_param;
} pthread_attr_t;

extern inline int pthread_attr_init(pthread_attr_t *attr)
{
	attr->stack_size = 3000;
	attr->sched_param.sched_priority = sched_get_priority_min(0);
	return 0;
}

extern inline int pthread_attr_getstacksize(pthread_attr_t *attr, size_t * stacksize)
{
	*stacksize = attr->stack_size;
	return 0;
}

extern inline int pthread_attr_setstacksize(pthread_attr_t *attr, size_t stacksize)
{
	attr->stack_size = stacksize;
	return 0;
}

extern inline void pthread_yield(void)
{
	rtl_pthread_suspend(pthread_self());
}

extern inline int pthread_setschedparam(pthread_t thread, int policy,
		    const struct rtl_sched_param *param) {
	thread->sched_param = *param;
	return 0;
}
 
extern inline int pthread_getschedparam(pthread_t thread, int *policy,
		    struct rtl_sched_param *param) {
	*param = thread->sched_param;
	return 0;
}

extern inline int pthread_attr_setschedparam(pthread_attr_t *attr,
		    const struct rtl_sched_param *param) {
	attr->sched_param = *param;
	return 0;
}

extern inline int pthread_attr_getschedparam(const pthread_attr_t *attr,
		    struct rtl_sched_param *param) {
	*param = attr->sched_param;
	return 0;
}
 


typedef void *(*RTL_START_ROUTINE)(void *);
extern int pthread_create (pthread_t *thread, pthread_attr_t *attr, void *(*start_routine)(void *), void *arg);

#define PTHREAD_CANCELED ((void *) -1)

extern void pthread_exit(void *retval);



/* end of POSIX interface */



/* unimplemented functions */

extern int rtl_pthread_bind_to_irq (pthread_t, unsigned int irq); /* need some sort of a RTL irq registry to implement this */

/* end of unimplemented functions */


#include <rtl_compat.h>

#endif
