#ifndef __RTL_ARCH_TIME_H__
#define __RTL_ARCH_TIME_H__

/*
 * rtl_time.h
 *
 * hardware clock interface
 *
 * Written by Michael Barabanov
 * Copyright VJY Associates
 * Released under the terms of the LGPL
 *
 */

#ifdef __KERNEL__
#include <linux/time.h>
#include <linux/tasks.h>
#include <asm/ptrace.h>
#include <asm/smp.h>

#ifdef __SMP__
#define rtl_get_cpu_id() hard_smp_processor_id()
#define rtl_num_cpus() (smp_num_cpus)
#else
#define rtl_get_cpu_id() 0
#define rtl_num_cpus() 1
#endif

static const struct timespec RTL_TIME_INFINITY = {0x7fffffff, 0};
static const struct timespec RTL_TIME_ZERO = {0, 0};

#define DECLARE_CPUID(cpu_id)   unsigned int cpu_id = rtl_get_cpu_id()

#include <linux/delay.h>

extern long rtl_delay_mult;

static inline void rtl_delay(long nanoseconds)
{
	__asm__("mull %0"
		:"=d" (nanoseconds)
		:"a" (nanoseconds),"0" (rtl_delay_mult)
		:"ax");
		
	__delay(nanoseconds);
}

struct clock_struct;
typedef int clockid_t;

typedef void (* RTL_CLOCK_HANDLER)(clockid_t, struct pt_regs *r);

struct clock_struct {
	int (*init) (clockid_t clock);
	void (*uninit) (clockid_t clock);
	int (*gettime) (clockid_t clock, struct timespec *ts);
	int (*settimer) (clockid_t clock, int mode, struct timespec *interval);
	void (*checklinuxirq) (clockid_t clock);
	RTL_CLOCK_HANDLER handler;
	int istimerset;
	int mode;
	struct timespec time;
	struct timespec resolution;
} __attribute__ ((aligned (64)));

#define RTL_NR_CLOCKS 5
extern struct clock_struct clocks [RTL_NR_CLOCKS];

#define clock_data(clock) (&clocks[clock])

/* POSIX interface */
#define CLOCK_REALTIME	0
#define CLOCK_NONE	1
#define CLOCK_8254	2
#define CLOCK_APIC	3 /* CLOCK_APIC+0, CLOCK_APIC+1, ...*/

#define clock_gettime(clock, tsptr) (clock_data(clock)->gettime(clock, tsptr))
#define clock_getres (clock, tsptr) { *(tsptr) = clock_data(clock)->resolution; }
/* end of POSIX interface */

extern clockid_t clock_local[NR_CPUS];

#define rtl_clock_local(cpu) (clock_local[cpu])

/* get the clock that generates interrupts for this clock */
#define rtl_clock_getirqclock(clock) \
	((clock) == CLOCK_REALTIME ? rtl_clock_local(rtl_get_cpu_id()) : clock)

/* modes */
#define CLOCK_MODE_UNINITIALIZED  -1
#define CLOCK_MODE_ONESHOT  0
#define CLOCK_MODE_PERIODIC 1

#define rtl_clock_settimer(clock, mode, tsptr) (clock_data(clock)->settimer(clock, mode, tsptr))
#define rtl_clock_getmode(clock) (clock_data(clock)->mode)


/* maximum interval between interrupts in nanoseconds */
#define RTL_MAX_INTERVAL (NSECS_PER_SEC/(HZ * 2))

/* scheduler interface */
extern int rtl_clock_sethandler (clockid_t clock, RTL_CLOCK_HANDLER handler);
extern int rtl_clock_unsethandler (clockid_t clock);
extern int rtl_clock_ishandlerset (clockid_t clock);
#define rtl_clock_checklinuxirq(clock) {clock_data(clock)->checklinuxirq(clock);}
/* end of scheduler interface */


/* for rtl_core */
#ifdef __SMP__
extern int request_ipi (unsigned int (*f)(struct pt_regs *r), int cpu);
extern int free_ipi (int cpu);
#endif



/* unimplemented functions */
extern int rtl_clock_init (clockid_t clock);
extern int rtl_clock_uninit (clockid_t clock);
extern int rtl_clock_getcpu (clockid_t clock); /* CPU this clock interrupts are passed to */
/* end of unimplemented functions list */


#endif

#define NSECS_PER_SEC 1000000000
#define timespec_add(t1, t2) { \
	(t1)->tv_nsec += (t2)->tv_nsec;  \
	(t1)->tv_sec += (t2)->tv_sec; \
	if ((t1)->tv_nsec >= NSECS_PER_SEC) { \
		(t1)->tv_nsec -= NSECS_PER_SEC; \
		(t1)->tv_sec++; \
	} else if ((t1)->tv_nsec < 0) { \
		(t1)->tv_nsec += NSECS_PER_SEC; \
		(t1)->tv_sec--; \
	} \
}

#define timespec_sub(t1, t2) { \
	(t1)->tv_nsec -= (t2)->tv_nsec;  \
	(t1)->tv_sec -= (t2)->tv_sec; \
	if ((t1)->tv_nsec < 0) { \
		(t1)->tv_nsec += NSECS_PER_SEC; \
		(t1)->tv_sec--; \
	} else if ((t1)->tv_nsec >= NSECS_PER_SEC) { \
		(t1)->tv_nsec -= NSECS_PER_SEC; \
		(t1)->tv_sec++; \
	} \
}

#define timespec_nz(t) ((t)->tv_sec != 0 || (t)->tv_nsec != 0)

#define timespec_lt(t1, t2) ((t1)->tv_sec < (t2)->tv_sec || ((t1)->tv_sec == (t2)->tv_sec && (t1)->tv_nsec < (t2)->tv_nsec))

#define timespec_gt(t1, t2) (timespec_lt(t2, t1))

#define timespec_ge(t1, t2) (!timespec_lt(t1, t2))

#define timespec_le(t1, t2) (!timespec_gt(t1, t2))

#define timespec_eq(t1, t2) ((t1)->tv_sec == (t2)->tv_sec && (t1)->tv_nsec == (t2)->tv_nsec)

static inline long long timespec_to_ns (struct timespec *ts)
{
	long long t;

	__asm__("imull %%edx\n"
		"add %%ebx, %%eax\n\t"
		"adc $0, %%edx\n\t"
		:"=A" (t)
		: "a" (ts->tv_sec), "d" (NSECS_PER_SEC), "b" (ts->tv_nsec)
		: "ax", "dx");
	return t;
}

static inline struct timespec timespec_from_ns (long long t)
{
	struct timespec ts;
	__asm__("idivl %%ecx\n\t"
		:"=a" (ts.tv_sec), "=d" (ts.tv_nsec)
		: "A" (t), "c" (NSECS_PER_SEC)
		:"cx");
	if (ts.tv_nsec < 0) {
		ts.tv_nsec += NSECS_PER_SEC;
		ts.tv_sec --;
	}
	return ts;
}


#endif

