/*
 * RT-Linux scheduler
 *
 * Written by Michael Barabanov, Victor Yodaiken
 * Copyright (C) VJY Associates LLC, 1998,1999
 * Released under the terms of the GNU Library General Public License
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/cons.h>
#include <asm/system.h>
#include <asm/segment.h>

#include <linux/timex.h>
#include <rtl_time.h>
#include <linux/rtl.h>
#include <rtl_sync.h>
#include <rtl_sched.h>

#include "x86/switch.h"

#define		rtl_make_unrtactive() { rtl_make_rt_system_idle();}
#define 	rtl_make_rtactive() {rtl_make_rt_system_active(); }
#define 	rtl_is_rtactive() (!rtl_rt_system_is_idle())

__asm__(".section .data\n\t"
	".align 64\n\t"
	".section .code\n\t");

/* struct timer_t rtl_timers [RTL_MAX_TIMERS]; */



struct rtl_sched_cpu_struct rtl_sched [NR_CPUS];

/* #define RTL_DEBUG */
int debugstate = 0;
int rtl_debug (int state)
{
	debugstate = state;
	return 0;
}

int rtl_startup(void *(*fn)(void *), void *data)
{
	rtl_allow_interrupts();

	sched_data(rtl_get_cpu_id()) -> rtl_current -> retval = (*fn)(data);
	rtl_pthread_delete (sched_data(rtl_get_cpu_id()) -> rtl_current);
	
	/* will never reach this line */
	return 0;
}

void pthread_exit(void *retval)
{
	sched_data(rtl_get_cpu_id()) -> rtl_current -> retval = retval;
	rtl_pthread_delete (pthread_self());
}

static void rtl_task_free_memory(void *p)
{
	RTL_THREAD *task = (RTL_THREAD *) p;
	kfree (task->stack_bottom);
	kfree (task);
}


	
static void add_to_task_list(pthread_t thread)
{
	DECLARE_CPUID(cpu_id);
	thread->next = sched_data(cpu_id) -> rtl_tasks;
	sched_data(cpu_id)->rtl_tasks = thread;
}

static int remove_from_this_cpu(pthread_t thread)
{
	int found = 0;
	RTL_THREAD *t;
	DECLARE_CPUID(cpu_id);
	if (thread != sched_data(thread->cpu)->rtl_tasks) {
		for (t = sched_data(thread->cpu)->rtl_tasks; t; t = t->next) {
			if (t->next == thread) {
				t->next = thread->next;
				found = 1;
				break;
			}
		}
		if (!found) {
			return -EINVAL;
		}
	} else {
		sched_data(thread->cpu)->rtl_tasks = thread->next;
	}

	if (sched_data(cpu_id)->rtl_task_fpu_owner == thread) {
		sched_data(cpu_id)->rtl_task_fpu_owner = 0;
	}

	return 0;
}

int timer_create(clockid_t clockid, void /*struct sigevent*/ *evp, timer_t *timerid)
{
	struct rtl_timer_struct * timer;
	timer = (struct rtl_timer_struct *) kmalloc (sizeof(struct rtl_timer_struct), GFP_KERNEL);
	if (!timer) {
		return -EAGAIN;
	}
	timer->spec.it_value = RTL_TIME_INFINITY;
	timer->spec.it_interval = RTL_TIME_ZERO;
	timer->clock = clockid;
	*timerid = timer;
	return 0;
}

int timer_delete(timer_t timerid)
{
	kfree(timerid);
	return 0;
}


/* TODO: handle ovalue */
int timer_settime(timer_t timerid, int flags,
		    const struct itimerspec *value, struct itimerspec *ovalue)
{
	if (!timespec_nz(&value->it_value)) {
		timerid->spec.it_value = RTL_TIME_INFINITY;
		timerid->spec.it_interval = RTL_TIME_ZERO;
		return 0;
	}
	timerid->spec = *value;
	if (!(flags & TIMER_ABSTIME)) {
		struct timespec ts;
		clock_gettime(timerid->clock, &ts);
		timespec_add(&timerid->spec.it_value, &ts);  /* so the spec.it_value contains the
							      absolute time of timer expiration */
	}
	return 0;
}


/* TODO */
int timer_gettime(timer_t timerid, struct itimerspec *value)
{
	return -EINVAL;
}

/* TODO */
int timer_getoverrun(timer_t timerid)
{
	return -EINVAL;
}


#ifdef __SMP__

static void rtl_task_move_on_cpu(void *p)
{
	RTL_THREAD *thread = (RTL_THREAD *) p;
	DECLARE_CPUID(cpu_id);
/* 	if (thread->clock == rtl_clock_local(thread->cpu)) { */
/* 		thread->clock = rtl_clock_local(cpu_id); */
/* 	} */
	thread->cpu = cpu_id;
	add_to_task_list (thread);
}

#endif


/* #define RTL_SCHED_DEBUG */

void rtl_schedule (void)
{
	RTL_THREAD *t;
	DECLARE_CPUID(cpu_id);
	RTL_THREAD *newt = &sched_data(cpu_id)->rtl_linux_task;
	RTL_THREAD *nextt = 0; /* task that will preempt, if any */
	unsigned long interrupt_state;

	struct timespec ts;

	rtl_no_interrupts(interrupt_state);

	/* expire timers */
	for (t = sched_data(cpu_id) -> rtl_tasks; t; t = t->next) {
		if (t->state == RTL_THREAD_DELAYED) {
#ifdef RTL_DEBUG
			if (debugstate & RTL_DEBUG_TIMER) {
/* 				conpr("d"); */
			}
#endif
			clock_gettime (t->timer->clock, &ts);
			if (timespec_ge (&ts, &t->timer->spec.it_value)) {
#ifdef RTL_DEBUG
				if (debugstate & RTL_DEBUG_TIMER) {
/* 					conpr("r"); */
				}
#endif
				t->state = RTL_THREAD_READY;
				if (timespec_nz(&t->timer->spec.it_interval)) { /* periodic */
					timespec_add(&t->timer->spec.it_value, &t->timer->spec.it_interval);
				} else {
					t->timer->spec.it_value = RTL_TIME_INFINITY;
				}
			} else {
				/*
				if (debugstate & RTL_DEBUG_TIMER) {
					conprtime(ts);
					conprtime(t->timer->spec.it_value);
				}
				*/
			}
		}
	}


        /* find the task to run  */
	for (t = sched_data(cpu_id)->rtl_tasks; t; t = t->next) {
		if (t->state == RTL_THREAD_READY) {
			if (t->sched_param.sched_priority > newt->sched_param.sched_priority) {
				newt = t;
			}
		}
	}


	if (rtl_clock_getmode(rtl_clock_local(cpu_id)) == CLOCK_MODE_ONESHOT) {
		struct timespec ts;
		struct timespec ts2;
		for (t = sched_data(cpu_id)->rtl_tasks; t; t = t->next) {
			if (t->state == RTL_THREAD_DELAYED && rtl_clock_getirqclock(t->timer->clock) == rtl_clock_local(cpu_id)) {
				if (t->sched_param.sched_priority > newt->sched_param.sched_priority && (!nextt ||(timespec_lt (&t->timer->spec.it_value, &nextt->timer->spec.it_value)))) {
				nextt = t;
				}
			}
		}


		if (nextt) {
			ts2 = nextt->timer->spec.it_value;
			clock_gettime (CLOCK_REALTIME, &ts);
			timespec_sub (&ts2, &ts);
			rtl_clock_settimer (rtl_clock_local(cpu_id), CLOCK_MODE_ONESHOT, &ts2); 
		}
	}

	if (newt ==  &sched_data(cpu_id)->rtl_linux_task) {
/* 		newt = &sched_data(cpu_id)->rtl_linux_task; */
		rtl_make_unrtactive();
	} else {
		rtl_make_rtactive();
	}

	if (newt != sched_data(cpu_id)->rtl_current) { /* switch out old, switch in new */

		rtl_switch_to(&sched_data(cpu_id)->rtl_current, newt);

		/* delay switching the FPU context until it is really needed */
		if (sched_data(cpu_id)->rtl_current -> uses_fp && sched_data(cpu_id)->rtl_task_fpu_owner != sched_data(cpu_id)->rtl_current) {
			if (sched_data(cpu_id)->rtl_task_fpu_owner) {
/* 				conpr("s"); */
				rtl_fpu_save (sched_data(cpu_id)->rtl_task_fpu_owner);
			}
/* 			conpr("r"); */

			rtl_fpu_restore (sched_data(cpu_id)->rtl_current);
			sched_data(cpu_id)->rtl_task_fpu_owner = sched_data(cpu_id)->rtl_current;
		}
	}
#ifdef RTL_SCHED_DEBUG
	{
		static struct timespec delay = {0, 25000};
		if (sched_data(cpu_id)->rtl_current !=  &sched_data(cpu_id)->rtl_linux_task) {
			struct timespec ts;
			clock_gettime (CLOCK_REALTIME, &ts);
			timespec_sub(&ts, &sched_data(cpu_id)->rtl_current->resume_time);
			if (timespec_gt(&ts, &delay)) {
	/* 			delay = ts; */
				conprtime(ts);
			}
		}
	}
#endif
	rtl_restore_interrupts(interrupt_state);
}


int pthread_create (pthread_t *thread, pthread_attr_t *attr, void *(*start_routine)(void *), void *arg)
{
	int *st;
	long interrupt_state;
	RTL_THREAD *task;
	pthread_attr_t default_attr;
	int stack_size;
	DECLARE_CPUID(cpu_id);

	if (!attr) {
		pthread_attr_init(&default_attr);
		attr = &default_attr;
	}
	stack_size = attr->stack_size;

	task = (RTL_THREAD *) kmalloc (sizeof(RTL_THREAD), GFP_KERNEL);
	if (!task) {
		return -ENOMEM;
	}

	st = (int *) kmalloc(stack_size, GFP_KERNEL);	
	if (!st) {
		kfree (task);
		return -ENOMEM;
	}

	*thread = task;

	task->being_destroyed = 0;
	task->retval = 0;

	task->free_task.next = 0;
	task->free_task.sync = 0;
	task->free_task.data = task;
	task->free_task.routine = rtl_task_free_memory;

	task->timer = 0;

#ifdef __SMP__
	task->move_task.next = 0;
	task->move_task.sync = 0;
	task->move_task.data = task;
	task->move_task.routine = rtl_task_move_on_cpu;
#endif

	task->sched_param = attr->sched_param;
	task->state = RTL_THREAD_READY; 
	task->stack_bottom = st;
	task->cpu = cpu_id;
	task->stack = st + stack_size / sizeof(int);
	task->uses_fp = 0;
	x86_init_stack(task,start_routine, arg,rtl_startup);
	rtl_no_interrupts(interrupt_state);
	add_to_task_list(task);
	rtl_restore_interrupts(interrupt_state);
	rtl_schedule();
	return 0;
}


int rtl_pthread_use_fp (pthread_t thread, int flag)
{
	DECLARE_CPUID(cpu_id);
	int flags;
	if (cpu_id != thread->cpu) {
		conpr("rtl_pthread_use_fp() called on a wrong CPU!\n");
		return -1;
	}
	rtl_no_interrupts(flags);
	thread -> uses_fp = flag;
	if (flag) {
/* 		conpr("initing"); */
		rtl_task_init_fpu (thread, sched_data(cpu_id)->rtl_task_fpu_owner);
		conpr("done");
	} else {
		if (sched_data(cpu_id)->rtl_task_fpu_owner == thread) {
			sched_data(cpu_id)->rtl_task_fpu_owner = 0;
		}
	}

	rtl_restore_interrupts(flags);
	return 0;
}


static void rtl_sched_timer_interrupt(clockid_t clock, struct pt_regs *regs)
{
	DECLARE_CPUID(cpu_id);
#ifdef RTL_DEBUG
	if (debugstate & RTL_DEBUG_HEARTBEAT) {
		conpr("h");
	}
#endif

#ifdef RTL_SCHED_DEBUG2
	static struct timespec delay = {0, 0};
#endif

#ifdef RTL_SCHED_DEBUG2
			{
				timespec_sub(&ts, &task->resume_time);
				if (timespec_gt(&ts, &delay)) {
					delay = ts;
					conprtime(delay);
					conprn(regs->eip);
					conpr("<== addr\n");
				}
			}
#endif
	rtl_schedule ();
	if (sched_data(cpu_id)->rtl_current == &sched_data(cpu_id)->rtl_linux_task) {
		rtl_clock_checklinuxirq (clock);
	}
}



int rtl_pthread_delete (pthread_t thread)
{
	RTL_THREAD *task = thread;
	int ret;
	long interrupt_state;

#ifdef __SMP__
	DECLARE_CPUID(cpu_id);
	if (cpu_id == task->cpu) {
#endif

		rtl_no_interrupts (interrupt_state);

		ret = remove_from_this_cpu (thread);
		if (ret < 0) {
			rtl_restore_interrupts (interrupt_state);
			return ret;
		}

		task->state = RTL_THREAD_ZOMBIE;

		queue_task(&task->free_task, &tq_scheduler);
		rtl_restore_interrupts (interrupt_state);
#ifdef __SMP__
	} else { /* the task is on the other CPU */
		rtl_no_interrupts (interrupt_state);
		set_bit (0, &task->being_destroyed);
		rtl_reschedule (task->cpu);
		rtl_restore_interrupts (interrupt_state);
		while (test_bit (0, &task->being_destroyed));
	}
#endif

	return 0;
}

static int cpu_exists (int cpu)
{
	int n;
	int i;
	for (i = 0; i < rtl_num_cpus(); i++) {
		n = cpu_logical_map (i);
		if (n == cpu) {
			return 1;
		}
	}
	return 0;
}


int rtl_pthread_assign_to_cpu (pthread_t thread, int cpu)
{
	int flags;
	DECLARE_CPUID(cpu_id);
	if (!cpu_exists(cpu)) {
		return -EINVAL;
	}
	if (thread->cpu != cpu_id) {
		return -EINVAL; /* can only move tasks from this CPU */
	}
	if (thread->cpu == cpu) {
		return 0; /* nothing to do */
	}

	rtl_no_interrupts(flags);
	remove_from_this_cpu (thread);
#ifdef __SMP__
	queue_task (&thread->move_task, &sched_data(cpu)->rtl_tq_cpu);
	rtl_reschedule (cpu);
#endif
	rtl_restore_interrupts(flags);
	/* should we mb() here? */
	rtl_schedule ();
	return 0;
}


int rtl_pthread_suspend(pthread_t thread)
{
	thread->state = RTL_THREAD_DORMANT;
	rtl_schedule();
	return 0;
}


int rtl_pthread_wakeup(pthread_t thread)
{
	thread->state = RTL_THREAD_READY;
	rtl_schedule();
	return 0;
}



int rtl_clock_set_periodic_mode (clockid_t clock, struct timespec *period)
{
	return rtl_clock_settimer (clock, CLOCK_MODE_PERIODIC, period);
}

int rtl_clock_set_oneshot_mode (clockid_t clock)
{
	int ret = rtl_clock_settimer (clock, CLOCK_MODE_ONESHOT, NULL);
	rtl_schedule();
	return ret;
}



int rtl_pthread_wait_timer (timer_t timer)
{
	long interrupt_state;
	DECLARE_CPUID(cpu_id);
	rtl_no_interrupts(interrupt_state);
	sched_data(cpu_id)->rtl_current->state = RTL_THREAD_DELAYED;
	sched_data(cpu_id)->rtl_current->timer = timer;
	rtl_schedule ();
	/* "signal is delivered"; here we could get the overrun count? */
	rtl_restore_interrupts(interrupt_state);
	return 0;
}


#ifdef __SMP__
unsigned int resched_irq(struct pt_regs *r)
{
	DECLARE_CPUID (cpu_id);
	RTL_THREAD *t;
	for (t = sched_data(cpu_id)->rtl_tasks; t; t = t->next) {
		if (test_bit(0, &t->being_destroyed)) {
			rtl_pthread_delete (t);
			clear_bit(0, &t->being_destroyed);
		}
	}
	run_task_queue (&sched_data(cpu_id)->rtl_tq_cpu);
	rtl_schedule();
	
	return 0;
}
#endif

int init_module(void)
{
	int interrupt_state;
	int cpu;
	int i;
	DECLARE_CPUID(cpu_id);

	rtl_no_interrupts(interrupt_state);

	for (i = 0; i < rtl_num_cpus(); i++) {
		cpu = cpu_logical_map (i);
#ifdef __SMP__
		sched_data(cpu) -> rtl_tq_cpu = NULL;
#endif
		sched_data(cpu) -> rtl_current = &sched_data(cpu)->rtl_linux_task;
		sched_data(cpu) -> rtl_tasks = &sched_data(cpu)->rtl_linux_task;

		sched_data(cpu) -> rtl_linux_task . state = RTL_THREAD_READY;
		sched_data(cpu) -> rtl_linux_task . sched_param . sched_priority = -1;
		sched_data(cpu) -> rtl_linux_task . next = &sched_data(cpu)->idle_task;
		sched_data(cpu) -> rtl_linux_task . uses_fp = 1;

		sched_data(cpu) -> idle_task . state = RTL_THREAD_READY;
		sched_data(cpu) -> idle_task . sched_param . sched_priority = -2;
		sched_data(cpu) -> idle_task . uses_fp = 0;
		sched_data(cpu) -> idle_task . next = 0;


		sched_data(cpu) -> rtl_task_fpu_owner = &sched_data(cpu)->rtl_linux_task;
		rtl_clock_sethandler (rtl_clock_local(cpu), rtl_sched_timer_interrupt);
	}

	rtl_clock_settimer(rtl_clock_local(cpu_id), CLOCK_MODE_ONESHOT, NULL);
#ifdef __SMP__

	for (i = 0; i < rtl_num_cpus(); i++) {
		int ret;
		cpu = cpu_logical_map (i);
		ret = request_ipi(resched_irq, cpu);
	}

#endif
	rtl_restore_interrupts (interrupt_state);
	return 0;
}


void cleanup_module(void)
{
	int i;
	int cpu;

	for (i = 0; i < rtl_num_cpus(); i++) {
		cpu = cpu_logical_map (i);
		rtl_clock_unsethandler (rtl_clock_local(cpu));
#ifdef __SMP__
		free_ipi(cpu);
#endif
	}
}


/* compatibility */
void *rtl_compat_start_routine(void *compat)
{
	void (*fn)(int data) = ((struct rtl_compat_struct *) compat) -> fn;
	int data = ((struct rtl_compat_struct *) compat) -> data;
	timer_t timer = ((struct rtl_compat_struct *) compat) -> timer;
	rtl_pthread_wait_timer(timer);
	conpr("s");
	fn(data);
	return NULL;
}

